"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var pouchdb_adapter_utils_1 = require("pouchdb-adapter-utils");
var pouchdb_merge_1 = require("pouchdb-merge");
var pouchdb_json_1 = require("pouchdb-json");
var pouchdb_errors_1 = require("pouchdb-errors");
var constants_1 = require("./constants");
var utils_1 = require("./utils");
function websqlBulkDocs(dbOpts, req, opts, api, db, websqlChanges, callback) {
    var newEdits = opts.new_edits;
    var userDocs = req.docs;
    var docInfos = userDocs.map(function (doc) {
        if (doc._id && pouchdb_adapter_utils_1.isLocalId(doc._id)) {
            return doc;
        }
        return pouchdb_adapter_utils_1.parseDoc(doc, newEdits, dbOpts);
    });
    var docInfoErrors = docInfos.filter(function (docInfo) {
        return docInfo.error;
    });
    if (docInfoErrors.length) {
        return callback(docInfoErrors[0]);
    }
    var tx;
    var results = new Array(docInfos.length);
    var fetchedDocs = new Map();
    var preconditionErrored;
    function complete() {
        if (preconditionErrored) {
            return callback(preconditionErrored);
        }
        websqlChanges.notify(api._name);
        callback(null, results);
    }
    function verifyAttachment(digest, callback) {
        var sql = 'SELECT count(*) as cnt FROM ' + constants_1.ATTACH_STORE + ' WHERE digest=?';
        tx.executeSql(sql, [digest], function (tx, result) {
            if (result.rows.item(0).cnt === 0) {
                var err = pouchdb_errors_1.createError(pouchdb_errors_1.MISSING_STUB, 'unknown stub attachment with digest ' + digest);
                callback(err);
            }
            else {
                callback();
            }
        });
    }
    function verifyAttachments(finish) {
        var digests = [];
        docInfos.forEach(function (docInfo) {
            if (docInfo.data && docInfo.data._attachments) {
                Object.keys(docInfo.data._attachments).forEach(function (filename) {
                    var att = docInfo.data._attachments[filename];
                    if (att.stub) {
                        digests.push(att.digest);
                    }
                });
            }
        });
        if (!digests.length) {
            return finish();
        }
        var numDone = 0;
        var err;
        function checkDone() {
            if (++numDone === digests.length) {
                finish(err);
            }
        }
        digests.forEach(function (digest) {
            verifyAttachment(digest, function (attErr) {
                if (attErr && !err) {
                    err = attErr;
                }
                checkDone();
            });
        });
    }
    function writeDoc(docInfo, winningRev, winningRevIsDeleted, newRevIsDeleted, isUpdate, delta, resultsIdx, callback) {
        var recv;
        var err;
        function finish() {
            var data = docInfo.data;
            var deletedInt = newRevIsDeleted ? 1 : 0;
            var id = data._id;
            var rev = data._rev;
            var json = utils_1.stringifyDoc(data);
            var sql = 'INSERT INTO ' + constants_1.BY_SEQ_STORE + ' (doc_id, rev, json, deleted) VALUES (?, ?, ?, ?);';
            var sqlArgs = [id, rev, json, deletedInt];
            function insertAttachmentMappings(seq, callback) {
                var attsAdded = 0;
                var attsToAdd = Object.keys(data._attachments || {});
                if (!attsToAdd.length) {
                    return callback();
                }
                function checkDone() {
                    if (++attsAdded === attsToAdd.length) {
                        callback();
                    }
                    return false;
                }
                function add(att) {
                    var sql = 'INSERT INTO ' + constants_1.ATTACH_AND_SEQ_STORE + ' (digest, seq) VALUES (?,?)';
                    var sqlArgs = [data._attachments[att].digest, seq];
                    tx.executeSql(sql, sqlArgs, checkDone, checkDone);
                }
                for (var _i = 0, attsToAdd_1 = attsToAdd; _i < attsToAdd_1.length; _i++) {
                    var item = attsToAdd_1[_i];
                    add(item);
                }
            }
            tx.executeSql(sql, sqlArgs, function (tx, result) {
                var seq = result.insertId;
                insertAttachmentMappings(seq, function () {
                    dataWritten(tx, seq);
                });
            }, function () {
                var fetchSql = utils_1.select('seq', constants_1.BY_SEQ_STORE, null, 'doc_id=? AND rev=?');
                tx.executeSql(fetchSql, [id, rev], function (tx, res) {
                    var seq = res.rows.item(0).seq;
                    var sql = 'UPDATE ' + constants_1.BY_SEQ_STORE + ' SET json=?, deleted=? WHERE doc_id=? AND rev=?;';
                    var sqlArgs = [json, deletedInt, id, rev];
                    tx.executeSql(sql, sqlArgs, function (tx) {
                        insertAttachmentMappings(seq, function () {
                            dataWritten(tx, seq);
                        });
                    });
                });
                return false;
            });
        }
        var attachments = Object.keys(docInfo.data._attachments || {});
        function collectResults(attachmentErr) {
            if (!err) {
                if (attachmentErr) {
                    err = attachmentErr;
                    callback(err);
                }
                else if (recv === attachments.length) {
                    finish();
                }
            }
        }
        err = null;
        recv = 0;
        docInfo.data._id = docInfo.metadata.id;
        docInfo.data._rev = docInfo.metadata.rev;
        if (newRevIsDeleted) {
            docInfo.data._deleted = true;
        }
        function attachmentSaved(err) {
            recv++;
            collectResults(err);
        }
        attachments.forEach(function (key) {
            var att = docInfo.data._attachments[key];
            if (!att.stub) {
                var data = att.data;
                delete att.data;
                att.revpos = parseInt(winningRev, 10);
                var digest = att.digest;
                saveAttachment(digest, data, attachmentSaved);
            }
            else {
                recv++;
                collectResults();
            }
        });
        if (!attachments.length) {
            finish();
        }
        function dataWritten(tx, seq) {
            var id = docInfo.metadata.id;
            var revsToCompact = docInfo.stemmedRevs || [];
            if (isUpdate && api.auto_compaction) {
                revsToCompact = pouchdb_merge_1.compactTree(docInfo.metadata).concat(revsToCompact);
            }
            if (revsToCompact.length) {
                utils_1.compactRevs(revsToCompact, id, tx);
            }
            docInfo.metadata.seq = seq;
            var rev = docInfo.metadata.rev;
            delete docInfo.metadata.rev;
            var sql = isUpdate
                ? 'UPDATE ' +
                    constants_1.DOC_STORE +
                    ' SET json=?, max_seq=?, winningseq=' +
                    '(SELECT seq FROM ' +
                    constants_1.BY_SEQ_STORE +
                    ' WHERE doc_id=' +
                    constants_1.DOC_STORE +
                    '.id AND rev=?) WHERE id=?'
                : 'INSERT INTO ' + constants_1.DOC_STORE + ' (id, winningseq, max_seq, json) VALUES (?,?,?,?);';
            var metadataStr = pouchdb_json_1.safeJsonStringify(docInfo.metadata);
            var params = isUpdate ? [metadataStr, seq, winningRev, id] : [id, seq, seq, metadataStr];
            tx.executeSql(sql, params, function () {
                results[resultsIdx] = {
                    ok: true,
                    id: docInfo.metadata.id,
                    rev: rev,
                };
                fetchedDocs.set(id, docInfo.metadata);
                callback();
            });
        }
    }
    function websqlProcessDocs() {
        pouchdb_adapter_utils_1.processDocs(dbOpts.revs_limit, docInfos, api, fetchedDocs, tx, results, writeDoc, opts);
    }
    function fetchExistingDocs(callback) {
        if (!docInfos.length) {
            return callback();
        }
        var numFetched = 0;
        function checkDone() {
            if (++numFetched === docInfos.length) {
                callback();
            }
        }
        docInfos.forEach(function (docInfo) {
            if (docInfo._id && pouchdb_adapter_utils_1.isLocalId(docInfo._id)) {
                return checkDone();
            }
            var id = docInfo.metadata.id;
            tx.executeSql('SELECT json FROM ' + constants_1.DOC_STORE + ' WHERE id = ?', [id], function (tx, result) {
                if (result.rows.length) {
                    var metadata = pouchdb_json_1.safeJsonParse(result.rows.item(0).json);
                    fetchedDocs.set(id, metadata);
                }
                checkDone();
            });
        });
    }
    function saveAttachment(digest, data, callback) {
        var sql = 'SELECT digest FROM ' + constants_1.ATTACH_STORE + ' WHERE digest=?';
        tx.executeSql(sql, [digest], function (tx, result) {
            if (result.rows.length) {
                return callback();
            }
            sql = 'INSERT INTO ' + constants_1.ATTACH_STORE + ' (digest, body, escaped) VALUES (?,?,1)';
            tx.executeSql(sql, [digest, utils_1.escapeBlob(data)], function () {
                callback();
            }, function () {
                callback();
                return false;
            });
        });
    }
    pouchdb_adapter_utils_1.preprocessAttachments(docInfos, 'binary', function (err) {
        if (err) {
            return callback(err);
        }
        db.transaction(function (txn) {
            tx = txn;
            verifyAttachments(function (err) {
                if (err) {
                    preconditionErrored = err;
                }
                else {
                    fetchExistingDocs(websqlProcessDocs);
                }
            });
        }, utils_1.websqlError(callback), complete);
    });
}
exports.default = websqlBulkDocs;
//# sourceMappingURL=bulkDocs.js.map