const cachedDatabases = new Map();

// openDatabase passed in through opts (e.g. for node-websql)
function openDatabaseWithOpts(opts): Database {
  return opts.websql(opts.name, opts.version, opts.description, opts.size);
}

function openDBSafely(opts): { db?: Database; error?: any } {
  try {
    return {
      db: openDatabaseWithOpts(opts),
    };
  } catch (err) {
    return {
      error: err,
    };
  }
}

function openDB(opts): { db?: Database; error?: any } {
  let cachedResult = cachedDatabases.get(opts.name);

  if (!cachedResult) {
    cachedResult = openDBSafely(opts);
    cachedDatabases.set(opts.name, cachedResult);
  }

  return cachedResult;
}

export default openDB;
